import * as http from "http";
import * as fs from "fs";
export function upload(url, data) {
    return new Promise((resolve, reject) => {
        const req = http.request(url, {
            method: "PUT",
            headers: {
                "Content-Type": "application/octet-stream",
                "Content-Length": data.length,
            },
        }, (res) => {
            if (res.statusCode == 200)
                resolve(res.statusCode);
            else
                reject(new Error(`HTTP status code: ${res.statusCode}`));
        });
        req.on("error", (e) => {
            reject(new Error(`HTTP error: ${e.message}`));
        });
        req.end(data);
    });
}
export function ll_to_wave(data, channels) {
    if (channels < 1 || channels > 16)
        throw Error("Invalid channel number");
    const samples = Math.floor(data.length / (7 * 8));
    const audio = Buffer.alloc(samples * channels * 3);
    for (let s = 0; s < samples; ++s) {
        const data_off = s * (7 * 8);
        const audio_off = s * channels * 3;
        const read_sample = (c) => {
            if ((c & 1) == 0) {
                const b0 = data.readUInt8(data_off + (c >> 1) * 7 + 0);
                const b1 = data.readUInt8(data_off + (c >> 1) * 7 + 1);
                const b2 = data.readUInt8(data_off + (c >> 1) * 7 + 2);
                return b0 | (b1 << 8) | (b2 << 16);
            }
            else {
                const b0 = data.readUInt8(data_off + (c >> 1) * 7 + 3);
                const b1 = data.readUInt8(data_off + (c >> 1) * 7 + 4);
                const b2 = data.readUInt8(data_off + (c >> 1) * 7 + 5);
                const b3 = data.readUInt8(data_off + (c >> 1) * 7 + 6);
                return (b0 >> 4) | (b1 << 4) | (b2 << 12) | ((b3 & 0x0f) << 20);
            }
        };
        for (let c = 0; c < channels; ++c) {
            const sample = read_sample(c);
            audio.writeUInt8(sample & 0xff, audio_off + c * 3 + 0);
            audio.writeUInt8((sample >> 8) & 0xff, audio_off + c * 3 + 1);
            audio.writeUInt8((sample >> 16) & 0xff, audio_off + c * 3 + 2);
        }
    }
    return audio;
}
export function normalize_l16(s) {
    return s / 32768.0;
}
export function normalize_l24(s) {
    return s / 8388608.0;
}
export function read_wave_file(b) {
    const u8 = new Uint8Array(b);
    function get_fourcc(off) {
        return String.fromCharCode(u8[off], u8[off + 1], u8[off + 2], u8[off + 3]);
    }
    function get_uint32(off) {
        return u8[off] | (u8[off + 1] << 8) | (u8[off + 2] << 16) | (u8[off + 3] << 24);
    }
    function get_uint16(off) {
        return u8[off] | (u8[off + 1] << 8);
    }
    function get_guid(off) {
        const p0 = get_uint32(off);
        const p1 = get_uint16(off + 4);
        const p2 = get_uint16(off + 6);
        function h(v, width) {
            return v.toString(16).padStart(width, '0');
        }
        function b(a) {
            return h(u8[off + a], 2);
        }
        return `${h(p0, 8)}-${h(p1, 4)}-${h(p2, 4)}-${b(8)}${b(9)}-${b(10)}${b(11)}${b(12)}${b(13)}${b(14)}${b(15)}`;
    }
    if (u8.length < 20 || get_fourcc(0) !== "RIFF" || get_fourcc(8) !== "WAVE")
        throw new Error("Not a WAV file");
    let raw_hdr = null;
    let chunk_offset = 12;
    while (chunk_offset <= u8.length - 8) {
        const chunk_data_offset = chunk_offset + 8;
        const remaining_len = u8.length - chunk_data_offset;
        const chunk_type = get_fourcc(chunk_offset);
        const chunk_len = get_uint32(chunk_offset + 4);
        switch (chunk_type) {
            case "fmt ":
                if (raw_hdr !== null)
                    throw new Error("Not a valid WAV file: multiple 'fmt ' chunks");
                if (chunk_len > remaining_len)
                    throw new Error("Not a valid WAV file: the 'fmt ' chunk is cut off");
                if (chunk_len >= 16) {
                    let raw = {
                        fmt_tag: get_uint16(chunk_data_offset),
                        channels: get_uint16(chunk_data_offset + 2),
                        sample_rate: get_uint32(chunk_data_offset + 4),
                        bytes_per_second: get_uint32(chunk_data_offset + 8),
                        block_align: get_uint16(chunk_data_offset + 12),
                        bits_per_sample: get_uint16(chunk_data_offset + 14)
                    };
                    const bytes_per_sample = (raw.bits_per_sample + 7) >> 3;
                    if (raw.bytes_per_second != raw.block_align * raw.sample_rate ||
                        raw.block_align != raw.channels * bytes_per_sample ||
                        raw.channels == 0 || raw.channels > 80)
                        throw new Error(`Not a valid WAV file: format header values don't agree: ${JSON.stringify(raw)}`);
                    if (raw.fmt_tag == 65534) {
                        if (chunk_len < 16 + 24)
                            throw new Error("Not a valid WAV file: EXTENSIBLE format tag needs at least 40 bytes for the 'fmt ' chunk");
                        const ext = {
                            len: get_uint16(chunk_data_offset + 16),
                            valid_bits_per_sample: get_uint16(chunk_data_offset + 18),
                            channel_mask: get_uint32(chunk_data_offset + 20),
                            sub_format_guid: get_guid(chunk_data_offset + 24)
                        };
                        if (ext.len < 22)
                            throw new Error("Not a valid WAV file: Extension size is too small");
                        if (ext.len > chunk_len - 18)
                            throw new Error("Not a valid WAV file: Extension size is too large");
                        switch (ext.sub_format_guid) {
                            case "00000001-0000-0010-8000-00aa00389b71":
                                raw.fmt_tag = 1;
                                break;
                            case "00000003-0000-0010-8000-00aa00389b71":
                                raw.fmt_tag = 3;
                                break;
                            default:
                                throw new Error(`Unsupported subformat GUID: ${ext.sub_format_guid}`);
                        }
                    }
                    raw_hdr = {
                        block_align: raw.block_align,
                        num_channels: raw.channels,
                        sample_rate: raw.sample_rate,
                        sample_format: (() => {
                            if (raw.fmt_tag == 3 && raw.bits_per_sample == 32)
                                return "Float";
                            else if (raw.fmt_tag == 1 && raw.bits_per_sample == 16)
                                return "L16";
                            else if (raw.fmt_tag == 1 && raw.bits_per_sample == 24)
                                return "L24";
                            else if (raw.fmt_tag == 1 && raw.bits_per_sample == 32)
                                return "L32";
                            else
                                throw new Error(`Unsupported audio format: ${JSON.stringify(raw)}`);
                        })()
                    };
                }
                else
                    throw new Error("Not a valid WAV file: could not parse 'fmt ' header");
                break;
            case "data":
                if (raw_hdr == null)
                    throw new Error("Not a valid WAV file: require 'fmt ' chunk before 'data'");
                const hdr = {
                    num_samples: Math.floor(Math.min(remaining_len, chunk_len) / raw_hdr.block_align),
                    ...raw_hdr
                };
                const data_len = hdr.block_align * hdr.num_samples;
                switch (hdr.sample_format) {
                    case "L16": {
                        const data = new DataView(b, chunk_data_offset, data_len);
                        function read_l16(sample) {
                            const offset = sample * hdr.num_channels * 2;
                            return Array.from({ length: hdr.num_channels }, (_, idx) => data.getInt16(offset + idx * 2, true));
                        }
                        return {
                            hdr,
                            read_raw: (pos) => Array.from(read_l16(pos)),
                            read_normalized: (pos) => Array.from(read_l16(pos), normalize_l16)
                        };
                    }
                    case "L24": {
                        const data = new Uint8Array(b, chunk_data_offset, data_len);
                        function read_l24(sample) {
                            const offset = sample * hdr.num_channels * 3;
                            return Array.from({ length: hdr.num_channels }, (_, idx) => {
                                const off = offset + idx * 3;
                                return ((data[off] << 8) | (data[off + 1] << 16) | (data[off + 2] << 24)) >> 8;
                            });
                        }
                        return {
                            hdr,
                            read_raw: read_l24,
                            read_normalized: (pos) => read_l24(pos).map(normalize_l24)
                        };
                    }
                    case "L32": {
                        const data = new DataView(b, chunk_data_offset, data_len);
                        function read_l32(sample) {
                            const offset = sample * hdr.num_channels * 4;
                            return Array.from({ length: hdr.num_channels }, (_, idx) => data.getInt32(offset + idx * 4, true));
                        }
                        return {
                            hdr,
                            read_raw: (pos) => Array.from(read_l32(pos)),
                            read_normalized: (pos) => Array.from(read_l32(pos), (s) => s / 2147483648.0)
                        };
                    }
                    case "Float": {
                        const data = new DataView(b, chunk_data_offset, data_len);
                        function read_float(sample) {
                            const offset = sample * hdr.num_channels * 4;
                            return Array.from({ length: hdr.num_channels }, (_, idx) => data.getFloat32(offset + idx * 4, true));
                        }
                        return {
                            hdr,
                            read_raw: read_float,
                            read_normalized: read_float
                        };
                    }
                }
        }
        chunk_offset = chunk_data_offset + chunk_len + (chunk_len & 1);
    }
    throw new Error("Invalid WAV file: missing 'data' chunk");
}
export function load_wave(path) {
    return read_wave_file(fs.readFileSync(path, null).buffer);
}
export function convert_wave_file(w, fmt) {
    function clamp(x, min, max) {
        return Math.max(min, Math.min(x, max));
    }
    if (w.hdr.sample_format == fmt)
        return w;
    const new_hdr = {
        ...w.hdr,
        sample_format: fmt
    };
    if (w.hdr.sample_format == "L16" && fmt == "L24") {
        return {
            hdr: new_hdr,
            read_normalized: w.read_normalized,
            read_raw: (pos) => w.read_raw(pos).map((s) => s << 8)
        };
    }
    return {
        hdr: new_hdr,
        read_normalized: w.read_normalized,
        read_raw: (() => {
            switch (fmt) {
                case "L16":
                    return (pos) => w.read_normalized(pos).map((s) => clamp(Math.round(s * 32768.0), -32768, 32767));
                case "L24":
                    return (pos) => w.read_normalized(pos).map((s) => clamp(Math.round(s * 8388608.0), -8388608, 8388607));
            }
        })()
    };
}
export function truncate_wave_file(w, max_samples) {
    return {
        ...w,
        hdr: {
            ...w.hdr,
            num_samples: Math.min(w.hdr.num_samples, max_samples)
        }
    };
}
export function wave_file_replace_sample_rate(w, sample_rate) {
    return {
        ...w,
        hdr: {
            ...w.hdr,
            sample_rate
        }
    };
}
const WAVE_HDR_LEN = 44;
export class WaveWriter {
    fd;
    pos;
    buf;
    constructor(path) {
        this.fd = fs.openSync(path, "w");
        this.pos = WAVE_HDR_LEN;
        this.buf = Buffer.alloc(80 * 4);
    }
    static get_bits_per_sample(sample_format) {
        switch (sample_format) {
            case "L16": return 16;
            case "L24": return 24;
            case "L32": return 32;
            case "Float": return 32;
        }
    }
    write_header(hdr) {
        const sample_format = hdr.sample_format;
        const bits_per_sample = WaveWriter.get_bits_per_sample(sample_format);
        const block_align = (bits_per_sample / 8) * hdr.num_channels;
        const bytes_per_second = hdr.sample_rate * block_align;
        const data_len = block_align * hdr.num_samples;
        const file_len = WAVE_HDR_LEN + data_len;
        const buf = this.buf;
        function set_fourcc(offset, fourcc) {
            for (let p = 0; p < 4; p++) {
                buf.writeUInt8(fourcc.charCodeAt(p), offset + p);
            }
        }
        set_fourcc(0, "RIFF");
        buf.writeUInt32LE(file_len - 8, 4);
        set_fourcc(8, "WAVE");
        set_fourcc(12, "fmt ");
        buf.writeUInt32LE(16, 16);
        buf.writeUInt16LE(sample_format == "Float" ? 3 : 1, 20);
        buf.writeUInt16LE(hdr.num_channels, 22);
        buf.writeUInt32LE(hdr.sample_rate, 24);
        buf.writeUInt32LE(bytes_per_second, 28);
        buf.writeUInt16LE(block_align, 32);
        buf.writeUInt16LE(bits_per_sample, 34);
        set_fourcc(36, "data");
        buf.writeUInt32LE(data_len, 40);
        fs.writeSync(this.fd, buf, 0, WAVE_HDR_LEN, 0);
    }
    write_data(w) {
        const buf = this.buf;
        const num_samples = w.hdr.num_samples;
        let pos = this.pos;
        switch (w.hdr.sample_format) {
            case "L16":
                for (let sample = 0; sample < num_samples; ++sample) {
                    const samples = w.read_raw(sample);
                    samples.forEach((v, idx) => buf.writeInt16LE(v, idx * 2));
                    pos += fs.writeSync(this.fd, buf, 0, samples.length * 2, pos);
                }
                break;
            case "L24":
                for (let sample = 0; sample < num_samples; ++sample) {
                    const samples = w.read_raw(sample);
                    samples.forEach((v, idx) => buf.writeIntLE(v, idx * 3, 3));
                    pos += fs.writeSync(this.fd, buf, 0, samples.length * 3, pos);
                }
                break;
            case "L32":
                for (let sample = 0; sample < num_samples; ++sample) {
                    const samples = w.read_raw(sample);
                    samples.forEach((v, idx) => buf.writeInt32LE(v, idx * 4));
                    pos += fs.writeSync(this.fd, buf, 0, samples.length * 4, pos);
                }
                break;
            case "Float":
                for (let sample = 0; sample < num_samples; ++sample) {
                    const samples = w.read_raw(sample);
                    samples.forEach((v, idx) => buf.writeFloatLE(v, idx * 4));
                    pos += fs.writeSync(this.fd, buf, 0, samples.length * 4, pos);
                }
                break;
        }
        this.pos = pos;
    }
    close() {
        fs.closeSync(this.fd);
    }
}
export function write_wave_file(path, w) {
    const ww = new WaveWriter(path);
    try {
        ww.write_header(w.hdr);
        ww.write_data(w);
    }
    finally {
        ww.close();
    }
}
;
const SAMPLE_ALIGNMENT = 7 * 16;
export function wave_file_to_biw(w) {
    if (w.hdr.num_channels > 80)
        throw new Error("Too many channels");
    const padded_samples = (() => {
        const remainder = w.hdr.num_samples % SAMPLE_ALIGNMENT;
        return w.hdr.num_samples + (remainder > 0 ? SAMPLE_ALIGNMENT - remainder : 0);
    })();
    const d = new Date();
    return {
        header: {
            Date: `${d.getDate()}/${d.getMonth() + 1}/${d.getFullYear()}`,
            Time: `${d.getHours()}:${d.getMinutes()}:${d.getSeconds()}`,
            Channels: w.hdr.num_channels,
            BitDepth: (w.hdr.sample_format == "L16") ? 16 : 24,
            SampleRate: w.hdr.sample_rate,
            SamplesPerChannel: padded_samples
        },
        wave: w
    };
}
export function encode_biw(b) {
    const w = convert_wave_file(b.wave, "L24");
    if (w.hdr.num_channels > 80)
        throw new Error("Too many channels");
    if (b.header.SamplesPerChannel % SAMPLE_ALIGNMENT != 0)
        throw new Error(`Number of samples must be a multiple of ${SAMPLE_ALIGNMENT}`);
    const channel_status = new Uint8Array(24);
    channel_status[0] = 0x85;
    channel_status[2] = 0x2c;
    channel_status[23] = 0x2b;
    const num_samples = w.hdr.num_samples;
    const num_channels = w.hdr.num_channels;
    if (num_samples > b.header.SamplesPerChannel ||
        num_channels != b.header.Channels)
        throw new Error("Invalid BIW structure");
    const channels16 = (num_channels + 15) >>> 4;
    const hdr_string = JSON.stringify(b.header);
    const hdr_extra = 8 - (hdr_string.length % 8);
    const data_len = channels16 * b.header.SamplesPerChannel * 56;
    const data_offset = hdr_string.length + hdr_extra;
    const buf = Buffer.alloc(data_offset + data_len);
    buf.fill(hdr_string, 0, hdr_string.length);
    for (let sample = 0; sample < num_samples; ++sample) {
        const sample_offset = data_offset + sample * channels16 * 56;
        const samples = w.read_raw(sample);
        const csb_idx = sample % 192;
        const cuvz = ((csb_idx == 0) ? 1 : 0) |
            ((channel_status[csb_idx >> 3] >> (csb_idx & 7)) & 1) << 3;
        function get_value(ch) {
            return (ch < samples.length) ? (samples[ch] & 0xFFFFFF) | (cuvz << 24) : 0;
        }
        for (let ch = 0; ch < num_channels; ch += 2) {
            const channel_offset = sample_offset + (ch >>> 1) * 7;
            const v0 = get_value(ch);
            const v1 = get_value(ch + 1);
            buf.writeUInt32LE((v0 | ((v1 & 15) << 28)) >>> 0, channel_offset);
            buf.writeUIntLE(v1 >> 4, channel_offset + 4, 3);
        }
    }
    return buf;
}
